# The MIT License (MIT)
#
# Copyright (c) 2013 The Weizmann Institute of Science.
# Copyright (c) 2018 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
# Copyright (c) 2018 Institute for Molecular Systems Biology,
# ETH Zurich, Switzerland.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


"""
Manage the training data for component contributions.

References:
-----------
.. [1] Alberty (2006)
.. [2] Maden (2000)
.. [3] Thauer (1977)
.. [4] Wagman (1982)
.. [5] Dolfing (1992)
.. [6] Dolfing (1994)
.. [7] CRC biochemistry (2010)
.. [8] Prince (1987)
.. [9] Thauer (1977)
.. [10] CRC biochemistry (2010)
.. [11] Alberty (2006)
.. [12] Deppenmeier (2008)
.. [13] Saeki (1985)
.. [14] Unden (1997)

"""
import logging
import warnings
from io import BytesIO
from typing import Iterable, List, Optional, Set, Tuple, Union

import numpy as np
import pandas as pd
import pkg_resources
from equilibrator_cache import FARADAY, Q_, Compound, CompoundCache, R, Reaction
from equilibrator_cache.exceptions import (
    MissingDissociationConstantsException,
    ParseException,
)
from equilibrator_cache.reaction import (
    create_stoichiometric_matrix_from_reactions,
)

from . import ZENODO_DOI_GROUPS, ZENODO_DOI_TRAINING_DATA
from .diskcache import get_cached_file


logger = logging.getLogger(__name__)


class TrainingData(object):
    """A base class for handling Training Data."""

    def __init__(
        self,
        ccache: CompoundCache,
        override_ionic_strength: Optional[Q_] = None,
        override_temperature: Optional[Q_] = None,
        override_p_mg: Optional[Q_] = None,
    ):
        """
        Create a TrainingData object.

        Parameters
        ----------
        ccache : CompoundCache
            The compound cache used for looking up structures and pKas.
        override_ionic_strength : bool, optional
            If provided, overrides all ionic_strength values given in the data
            files with this value.
        override_temperature : bool, optional
            If provided, overrides all temperature values given in the data
            files with this value.
        override_p_mg : bool, optional
            If provided, overrides all p_mg values given in the data
            files with this value.

        """
        self.ccache = ccache
        self.S = None  # a DataFrame containing the stoichiometric matrix

        # read the group summary table from Zenodo
        group_file = get_cached_file(ZENODO_DOI_GROUPS, "group_summary.csv")
        self.group_summary = pd.read_csv(group_file, index_col=0)

        logger.info("Gathering reaction data")
        (
            self.reaction_df,
            self._non_decomposable_compounds,
        ) = self.create_reaction_dataframe()

        for cpd in self._non_decomposable_compounds:
            if self.ccache.is_proton(cpd):
                raise ValueError("Protons are not allowed in the training data")

        for compound in self.decomposable_compounds:
            if compound.atom_bag is None and compound.inchi is None:
                warnings.warn(
                    f"Compound {compound} has neither an atom "
                    f"bag nor an InChI."
                )
            elif compound.atom_bag is None and compound.inchi is not None:
                warnings.warn(
                    f"Compound {compound} has no atom bag, but "
                    f"does have an InChI: {compound.inchi}."
                )

        logger.info("Ensure all the relevant columns are in the correct units")
        self.assert_data()

        logger.info("Balancing reactions with H2O and H+")
        self.balance_reactions()

        logger.info("Creating the training stoichiometric matrix")
        self.create_stoichiometric_matrix_from_reactions()

        logger.info("Applying reverse Legendre transform on dG' values")
        standard_dg = self._reverse_transform(
            override_ionic_strength=override_ionic_strength,
            override_temperature=override_temperature,
            override_p_mg=override_p_mg,
        )
        self.reaction_df = self.reaction_df.assign(
            **{"standard_dg": list(standard_dg)}
        )

    def parse_formula(self, formula: str) -> Reaction:
        """Parse a chemical formula to create a Reaction object using ccache."""
        return Reaction.parse_formula(self.ccache.get_compound, formula)

    def create_reaction_dataframe(self) -> Tuple[pd.DataFrame, Set[Compound]]:
        """Create the DataFrame containing all the reaction training data."""
        raise NotImplementedError()

    @property
    def stoichiometric_matrix(self) -> pd.DataFrame:
        """Get the stoichiometric matrix as a DataFrame.

        :return: a Pandas DataFrame
        """
        return self.S

    @property
    def compounds(self) -> List[Compound]:
        """Get the list of compounds.

        :return: A list of Compound objects according to their order in S.
        """
        if self.S is None:
            compounds = {
                c
                for r in self.reaction_df.reaction
                for c in r.sparse
                if not self.ccache.is_proton(c)
            }
            if not any(map(self.ccache.is_water, compounds)):
                compounds.add(self.ccache.water)
            return sorted(compounds)
        else:
            return self.S.index.tolist()

    @property
    def non_decomposable_compounds(self) -> Set[Compound]:
        """Get the set of non-decomposable compounds.

        :return: A Set of Compound objects
        """
        return self._non_decomposable_compounds

    @property
    def decomposable_compounds(self) -> Set[Compound]:
        """Get the set of decomposable compounds.

        :return: A Set of Compound objects
        """
        return set(self.compounds).difference(self._non_decomposable_compounds)

    @property
    def standard_dg(self) -> pd.Series:
        """Get the standard deltaGs according to the reaction order in S.

        :return: A Pandas Series of the standard detlaGs
        """
        return self.reaction_df.standard_dg

    @property
    def weight(self) -> pd.Series:
        """Get the weights according to the reaction order in S.

        :return:A Pandas Series of the weights
        """
        return self.reaction_df.weight

    def assert_data(self):
        """Ensure all the relevant columns are in the correct units."""
        for col, dim in [
            ("standard_dg_prime", "[energy]/[substance]"),
            ("ionic_strength", "[concentration]"),
            ("temperature", "[temperature]"),
            ("p_h", None),
            ("p_mg", None),
        ]:
            assert col in self.reaction_df.columns

            for v in self.reaction_df[col].values:
                assert v.check(dim)

    def _balance_reaction(self, rxn: Reaction) -> Union[Reaction, None]:
        """Try balancing the reaction with H2O and H+.

        :param rxn: A Reaction object
        :return: A balanced Reaction, or None
        """
        if rxn.is_balanced(ignore_atoms=()):
            # the reaction is already balanced, nothing to do
            return rxn

        if rxn.is_balanced(ignore_atoms=("H",)):
            # the reaction just needs to be balanced by protons
            return rxn.balance_with_compound(
                self.ccache.proton, ignore_atoms=()
            )

        if not rxn.is_balanced(ignore_atoms=("H", "O", "e-")):
            # this reaction cannot be balanced only by H2O and H+, because
            # elements other than e-, O, and H are not balanced
            warnings.warn("This reaction cannot be balanced: " f"{rxn}")
            return None
        else:
            # we need to first balance the O atoms using water, and then
            # balance the H atoms using protons
            rxn = rxn.balance_with_compound(
                self.ccache.water, ignore_atoms=("H",)
            )
            return rxn.balance_with_compound(
                self.ccache.proton, ignore_atoms=()
            )

    def balance_reactions(self) -> None:
        """Balance the reactions.

        Use the chemical formulas from the InChIs to verify that each and
        every reaction is balanced. If it can be easily fixed with protons
        and/or water, fix it. Otherwise, drop this reaction from the training
        data.
        """
        balanced_reactions = []
        for row in self.reaction_df.itertuples(index=True):
            if "balance" in self.reaction_df.columns and not row.balance:
                balanced_reactions.append(row.reaction)
            else:
                balanced_reactions.append(self._balance_reaction(row.reaction))

        self.reaction_df["reaction"] = balanced_reactions
        self.reaction_df = self.reaction_df[
            ~pd.isnull(self.reaction_df.reaction)
        ]

    def create_stoichiometric_matrix_from_reactions(self) -> None:
        """Create the stoichiometric matrix for training.

        Convert the list of reactions in sparse notation into a full
        stoichiometric matrix, where the rows (compounds) are in the same
        order as in 'compounds' and the columns match the reaction indices.
        """
        self.S = create_stoichiometric_matrix_from_reactions(
            self.reaction_df.reaction,
            self.ccache.is_proton,
            self.ccache.is_water,
            self.ccache.water,
        )
        self.S.columns = self.reaction_df.index

    def _reverse_transform(
        self,
        override_ionic_strength: Optional[Q_] = None,
        override_temperature: Optional[Q_] = None,
        override_p_mg: Optional[Q_] = None,
    ) -> Iterable[float]:
        """Reverse Legendre transform.

        Parameters
        ----------
        override_ionic_strength : bool, optional
            If provided, overrides all ionic_strength values given in the data
            files with this value.
        override_temperature : bool, optional
            If provided, overrides all temperature values given in the data
            files with this value.
        override_p_mg : bool, optional
            If provided, overrides all p_mg values given in the data
            files with this value.

        """
        _reaction_df = self.reaction_df.copy()
        if override_ionic_strength:
            assert override_ionic_strength.check("[concentration]")
            _reaction_df.ionic_strength = override_ionic_strength
        if override_temperature:
            assert override_temperature.check("[temperature]")
            _reaction_df.temperature = override_temperature
        if override_p_mg:
            assert override_p_mg.check("")
            _reaction_df.p_mg = override_p_mg

        for row in _reaction_df.itertuples():
            standard_dg = row.standard_dg_prime
            try:
                standard_dg -= row.reaction.transform(
                    p_h=row.p_h,
                    ionic_strength=row.ionic_strength,
                    temperature=row.temperature,
                    p_mg=row.p_mg,
                )
            except MissingDissociationConstantsException as e:
                logger.warning(f"Cannot reverse transform {row.reaction}: {e}")
            yield standard_dg


class ToyTrainingData(TrainingData):
    """Toy training data class."""

    def __init__(self, **kw):
        """Create the ToyTrainingData object."""
        super(ToyTrainingData, self).__init__(**kw)

    def create_reaction_dataframe(self) -> Tuple[pd.DataFrame, Set[Compound]]:
        """Create the DataFrame of reactions for training."""
        reaction_df = pd.read_csv(
            pkg_resources.resource_filename(
                "component_contribution.data", "toy_training_data.csv"
            )
        )

        # convert the formula strings into Reaction objects
        reaction_df.loc[:, "reaction"] = reaction_df.reaction.apply(
            self.parse_formula
        )

        # add units to the relevant columns:
        for col, unit in [
            ("standard_dg_prime", "kJ/mol"),
            ("ionic_strength", "M"),
            ("temperature", "K"),
            ("p_h", None),
            ("p_mg", None),
        ]:
            reaction_df[col] = reaction_df[col].apply(lambda x: Q_(x, unit))

        non_decomposable_compounds = set()
        return reaction_df, non_decomposable_compounds


class FullTrainingData(TrainingData):
    """Full training data class."""

    def __init__(self, ccache: CompoundCache, **kw):
        """Create the FullTrainingData object."""
        super(FullTrainingData, self).__init__(ccache=ccache, **kw)

    def create_reaction_dataframe(self) -> Tuple[pd.DataFrame, Set[Compound]]:
        """Create the DataFrame of reactions for training."""
        tecr_df = self.read_tecrdb(
            get_cached_file(ZENODO_DOI_TRAINING_DATA, "TECRDB.csv")
        )
        tecr_df["weight"] = 1.0

        formation_df, non_decomposable_compounds = self.read_formations(
            get_cached_file(
                ZENODO_DOI_TRAINING_DATA, "formation_energies_transformed.csv"
            )
        )
        formation_df["weight"] = 1.0

        redox_df, more_non_decomposable_compounds = self.read_redox(
            get_cached_file(ZENODO_DOI_TRAINING_DATA, "redox.csv")
        )
        redox_df["weight"] = 1.0
        non_decomposable_compounds.update(more_non_decomposable_compounds)

        reaction_df = pd.concat([tecr_df, formation_df, redox_df], sort=False)
        reaction_df.reset_index(drop=True, inplace=True)

        return reaction_df, non_decomposable_compounds

    def read_tecrdb(self, tecr_file: BytesIO) -> pd.DataFrame:
        """Load a data frame with information from the TECRdb (NIST).

        The component-contribution package distributes data tables with
        information on the 'thermodynamics of enzyme-catalyzed
        reactions'[1, 2]_ that are used as training data.

        :return: a Pandas DataFrame of the TECRDB reaction

        References
        ----------
        .. [1] Goldberg, Robert N., Yadu B. Tewari, and Talapady N. Bhat.
               “Thermodynamics of Enzyme-Catalyzed Reactions—a Database for
               Quantitative Biochemistry.” Bioinformatics 20, no. 16
               (November 1, 2004): 2874–77.
               https://doi.org/10.1093/bioinformatics/bth314.
        .. [2] http://xpdb.nist.gov/enzyme_thermodynamics/

        """
        tecr_df = pd.read_csv(tecr_file)

        # assume a default ionic strength of 0.25 M
        tecr_df.ionic_strength.fillna(0.25, inplace=True)

        # assume a default pMg of 14 (i.e. a negligible [Mg2+])
        tecr_df.p_mg.fillna(14, inplace=True)

        # remove rows with missing essential data
        tecr_df = tecr_df[
            ~pd.isnull(tecr_df[["K_prime", "temperature", "p_h"]]).any(axis=1)
        ]

        for col, unit in [
            ("K_prime", None),
            ("ionic_strength", "M"),
            ("temperature", "K"),
            ("p_h", None),
            ("p_mg", None),
        ]:
            tecr_df[col] = tecr_df[col].apply(lambda x: Q_(x, unit))

        # calculate the dG'0 from the Keq and T
        standard_dg_primes = [
            (-R * row.temperature * np.log(row.K_prime)).to("kJ/mol")
            for row in tecr_df.itertuples()
        ]
        tecr_df = tecr_df.assign(
            **{"standard_dg_prime": standard_dg_primes, "balance": True}
        )

        # parse the reaction
        tecr_df.loc[:, "reaction"] = tecr_df["reaction"].apply(
            self.parse_formula
        )

        # skip reactions that could not be parsed
        tecr_df = tecr_df[~pd.isnull(tecr_df["reaction"])]

        tecr_df.drop(
            ["url", "method", "K", "K_prime", "eval", "EC", "enzyme_name"],
            axis=1,
            inplace=True,
        )

        logger.debug(
            "Successfully added %d reactions from TECRDB" % tecr_df.shape[0]
        )
        return tecr_df

    def read_formations(
        self, formation_file: BytesIO
    ) -> Tuple[pd.DataFrame, Set[Compound]]:
        """Read the Formation Energy data from literature.

        All the data we have for formation energies from [1-6].

        :return: a 2-tuple of the Pandas DataFrame of reactions, and the set
        of compounds that do not decompose into groups.
        """
        formation_df = pd.read_csv(formation_file)
        compounds = formation_df["cid"].apply(self.ccache.get_compound)

        if pd.isnull(compounds).any():
            missing_cids = formation_df.loc[pd.isnull(compounds), "cid"]
            raise ParseException(
                "Cannot find some of the compounds in the "
                "cache: " + str(missing_cids)
            )

        compounds_that_do_not_decompose = set(
            compounds.loc[formation_df.decompose == 0].values
        )

        # skip compounds that have no formation energy (they are in the table
        # only in order to indicate something, for example that they should
        # not be decomposed)
        formation_df = formation_df[~pd.isnull(formation_df.standard_dg_prime)]

        for col, unit in [
            ("standard_dg_prime", "kJ/mol"),
            ("ionic_strength", "M"),
            ("temperature", "K"),
            ("p_h", None),
            ("p_mg", None),
        ]:
            formation_df[col] = formation_df[col].apply(lambda x: Q_(x, unit))

        formation_df = formation_df.assign(
            **{
                "reaction": compounds.apply(lambda c: Reaction({c: 1})),
                "balance": False,
                "description": formation_df["name"] + " formation",
            }
        )
        formation_df.drop(
            ["name", "cid", "remark", "decompose"], axis=1, inplace=True
        )

        logger.debug(
            "Successfully added %d formation energies" % formation_df.shape[0]
        )
        return formation_df, compounds_that_do_not_decompose

    def read_redox(
        self, redox_file: BytesIO
    ) -> Tuple[pd.DataFrame, Set[Compound]]:
        """Read the Reduction potential from literature.

        All the reduction potentials we have from [7-14].

        :return: a Pandas DataFrame of the redox reactions.
        """
        redox_df = pd.read_csv(redox_file)

        for col, unit in [
            ("standard_E_prime", "V"),
            ("ionic_strength", "M"),
            ("temperature", "K"),
            ("p_h", None),
            ("p_mg", None),
        ]:
            redox_df[col] = redox_df[col].apply(lambda x: Q_(x, unit))

        compounds_ox = redox_df["CID_ox"].apply(self.ccache.get_compound)
        compounds_red = redox_df["CID_red"].apply(self.ccache.get_compound)

        compounds_that_do_not_decompose = set(
            compounds_ox.loc[redox_df.decompose == 0].values
        )
        compounds_that_do_not_decompose.update(
            compounds_red.loc[redox_df.decompose == 0].values
        )

        reaction = [
            Reaction({c_ox: -1, c_red: 1})
            for c_ox, c_red in zip(compounds_ox, compounds_red)
        ]

        delta_nH = redox_df.nH_red - redox_df.nH_ox
        delta_charge = redox_df.charge_red - redox_df.charge_ox
        redox_df["delta_e"] = delta_nH - delta_charge

        standard_dg_primes = [
            (-FARADAY * row.standard_E_prime * row.delta_e).to("kJ/mol")
            for row in redox_df.itertuples()
        ]

        redox_df = redox_df.assign(
            **{
                "reaction": reaction,
                "description": redox_df["name"] + " redox",
                "standard_dg_prime": standard_dg_primes,
                "balance": False,
            }
        )
        redox_df.drop(
            [
                "name",
                "CID_ox",
                "CID_red",
                "charge_ox",
                "charge_red",
                "nH_ox",
                "nH_red",
                "standard_E_prime",
                "delta_e",
            ],
            axis=1,
            inplace=True,
        )

        logger.debug(
            "Successfully added %d redox potentials" % redox_df.shape[0]
        )
        return redox_df, compounds_that_do_not_decompose



"""wrapper for component-contribution predictions."""
# The MIT License (MIT)
#
# Copyright (c) 2013 The Weizmann Institute of Science.
# Copyright (c) 2018 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
# Copyright (c) 2018 Institute for Molecular Systems Biology,
# ETH Zurich, Switzerland.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import logging
import warnings
from typing import Dict, Iterable, List, Optional, Tuple

import numpy as np
from equilibrator_cache import FARADAY, Q_, Compound, R, Reaction, ureg
from equilibrator_cache.exceptions import MissingDissociationConstantsException
from equilibrator_cache.thermodynamic_constants import default_pMg

from . import CCModelParameters, Preprocessor


logger = logging.getLogger(__name__)


class GibbsEnergyPredictor(object):
    """A class that can be used to predict dGs of reactions using CC."""

    def __init__(
        self,
        parameters: CCModelParameters = None,
        rmse_inf: Optional[Q_] = None,
    ):
        """Create a GibbsEnergyPredictor object.

        parameters : CCModelParameters
            Optional CC parameters. If not provided, the parameters are
            automatically downloaded from Zenodo.
        rmse_inf : Qunatity
            another parameter determining the uncertainty we associate
            with reactions that are not covered at all (not by RC nor GC)
            (Default value = 1e5 kJ/mol)
        """
        self.params = parameters or CCModelParameters.from_zenodo()

        if rmse_inf is not None:
            assert rmse_inf.check(
                "[energy]/[substance]"
            ), "rmse_inf must be in kJ/mol or equivalent units"
            rmse_inf = rmse_inf.m_as("kJ/mol")
            self.preprocess = Preprocessor(self.params, rmse_inf=rmse_inf)
        else:
            self.preprocess = Preprocessor(self.params)

    def get_compound_prediction(
        self, compound: Compound
    ) -> Tuple[Optional[float], Optional[np.ndarray]]:
        """Get the (mu, sigma) predictions of a compound's formation energy."""
        return self.preprocess.get_compound_prediction(compound)

    def get_reaction_prediction(
        self, reaction: Reaction
    ) -> Tuple[float, np.ndarray, Dict[Compound, float]]:
        """Get the (mu, sigma) predictions of a reaction's energy.

        See the Preprocessor class for details.
        """
        return self.preprocess.get_reaction_prediction(reaction)

    def get_reaction_prediction_multi(
        self, reactions: List[Reaction]
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Get the (mu, sigma) predictions of a reaction's energy.

        See the Preprocessor class for details.
        """
        return self.preprocess.get_reaction_prediction_multi(reactions)

    def standard_dgf(self, compound: Compound) -> ureg.Measurement:
        """Calculate the chemical formation energy of the major MS at pH 7.

        Parameters
        ----------
        compound: Compound :
            a compound object


        Returns
        -------
        Measurement
            a tuple of two arrays. the first is a 1D NumPy array
            containing the CC estimates for the reactions' untransformed dG0
            (i.e. using the major MS at pH 7 for each of the reactants).
            the second is a 2D numpy array containing the covariance matrix
            of the standard errors of the estimates. one can use the
            eigenvectors of the matrix to define a confidence high-dimensional
            space, or use U as the covariance of a Gaussian used for sampling
            (where dG0_cc is the mean of that Gaussian).

        """
        mu, sigma = self.get_compound_prediction(compound)
        if mu is None:
            return Q_(0, "kJ/mol").plus_minus(self.preprocess.RMSE_inf)
        else:
            return Q_(mu, "kJ/mol").plus_minus(np.linalg.norm(sigma, 2))

    def standard_dg(self, reaction: Reaction) -> ureg.Measurement:
        r"""Calculate the chemical reaction energy.

        Using the major microspecies of each of the reactants.

        Parameters
        ----------
        reaction: Reaction :
            the input Reaction object


        Returns
        -------
        Measurement
        standard_dg : Measurement
            the :math:`\Delta G` in kJ/mol and standard error. to
            calculate the 95% confidence interval, multiply the error by 1.96

        """
        mu, sigma, residual = self.get_reaction_prediction(reaction)
        if residual:
            return Q_(0, "kJ/mol").plus_minus(self.preprocess.RMSE_inf)
        else:
            return Q_(mu, "kJ/mol").plus_minus(np.linalg.norm(sigma, 2))

    def standard_dg_multi(
        self, reactions: List[Reaction]
    ) -> Tuple[np.ndarray, np.ndarray]:
        r"""Calculate the chemical reaction energies for a list of reactions.

        Using the major microspecies of each of the reactants.

        Parameters
        ----------
        reactions : List[Reaction]
            a list of Reaction objects


        Returns
        -------
        standard_dg : np.ndarray
            the array of Component Contribution estimates for the reactions'
            untransformed :math:`\Delta G` in kJ/mol,
            (i.e. using the major MS at pH 7 for each of the reactants).
        cov_dg : np.ndarray
            the covariance matrix of the standard errors of the estimates.
            one can use the eigenvectors of the matrix to define a confidence
            high-dimensional space, or use U as the covariance of a Gaussian
            used for sampling (where cov_dg is the mean of that Gaussian).

        """
        mu, sigma, residual = self.get_reaction_prediction_multi(reactions)
        standard_dg = Q_(mu, "kJ/mol")
        cov_dg = Q_(sigma @ sigma.T + residual.T @ residual, "(kJ/mol)**2")
        return standard_dg, cov_dg

    @ureg.check(None, None, "", "[concentration]", "[temperature]", "")
    def standard_dgf_prime(
        self,
        compound: Compound,
        p_h: Q_,
        ionic_strength: Q_,
        temperature: Q_,
        p_mg: Q_ = default_pMg,
    ) -> ureg.Measurement:
        """Calculate the biocheimcal formation energy of the compound.

        Parameters
        ----------
        compound : Compound
            a compound object
        p_h : Quantity
            the pH
        ionic_strength : Quantity
            the ionic strength
        temperature : Quantity
            temperature in Kalvin
        p_mg : Quantity, optional
            the pMg (Default value = default_pMg)


        Returns
        -------
        type
            a tuple of two arrays. the first is a 1D NumPy array
            containing the CC estimates for the reactions' untransformed dG0
            (i.e. using the major MS at pH 7 for each of the reactants).
            the second is a 2D numpy array containing the covariance matrix
            of the standard errors of the estimates. one can use the
            eigenvectors of the matrix to define a confidence high-dimensional
            space, or use U as the covariance of a Gaussian used for sampling
            (where dG0_cc is the mean of that Gaussian).

        """
        return self.standard_dgf(compound) + compound.transform(
            p_h=p_h,
            ionic_strength=ionic_strength,
            temperature=temperature,
            p_mg=p_mg,
        )

    @ureg.check(None, None, "", "[concentration]", "[temperature]", "")
    def standard_dg_prime(
        self,
        reaction: Reaction,
        p_h: Q_,
        ionic_strength: Q_,
        temperature: Q_,
        p_mg: Q_ = default_pMg,
    ) -> ureg.Measurement:
        r"""Calculate the transformed reaction energies of a reaction.

        Parameters
        ----------
        reaction : Reaction
            the input Reaction object
        p_h : Q_
            pH
        ionic_strength : Q_
            ionic strength
        temperature : Q_
            temperature
        p_mg: Q_ :
             (Default value = default_pMg)

        Returns
        -------
        standard_dg : Measurement
            the :math:`\Delta G'` in kJ/mol and standard error. to
            calculate the 95% confidence interval, multiply the error by 1.96

        """
        return self.standard_dg(reaction) + reaction.transform(
            p_h=p_h,
            ionic_strength=ionic_strength,
            temperature=temperature,
            p_mg=p_mg,
        )

    @ureg.check(None, None, "", "[concentration]", "[temperature]", "")
    def standard_dg_prime_multi(
        self,
        reactions: List[Reaction],
        p_h: Q_,
        ionic_strength: Q_,
        temperature: Q_,
        p_mg: Q_ = default_pMg,
    ) -> Tuple[np.ndarray, np.ndarray]:
        r"""Calculate the transformed reaction energies of a list of reactions.

        Parameters
        ----------
        reactions : List[Reaction]
            a list of Reaction objects
        p_h : Quantity
            the pH
        ionic_strength : Quantity
            the ionic strength
        temperature : Quantity
            the temperature
        p_mg: Quantity, optional
            (Default value = default_pMg)

        Returns
        -------
        standard_dg_prime : np.ndarray
            the array of Component Contribution estimates for the reactions'
            :math:`\Delta G'` in kJ/mol
        cov_dg : np.ndarray
            the covariance matrix of the standard errors of the estimates.
            one can use the eigenvectors of the matrix to define a confidence
            high-dimensional space, or use U as the covariance of a Gaussian
            used for sampling (where cov_dg is the mean of that Gaussian).

        """
        standard_dg, cov_dg = self.standard_dg_multi(reactions)

        for i, r in enumerate(reactions):
            try:
                standard_dg[i] += r.transform(
                    p_h=p_h,
                    ionic_strength=ionic_strength,
                    temperature=temperature,
                    p_mg=p_mg,
                )
            except MissingDissociationConstantsException as e:
                warnings.warn(
                    f"Cannot calculate Legendre transform for reaction #{i}: "
                    + str(e)
                )

        return standard_dg, cov_dg

    @ureg.check(
        None,
        None,
        None,
        None,
        None,
        "",
        "",
        "[concentration]",
        "[concentration]",
        "[energy]/[current]/[time]",
        "[temperature]",
        "",
        "",
    )
    def multicompartmental_standard_dg_prime(
        self,
        reaction_1: Reaction,
        reaction_2: Reaction,
        transported_protons: float,
        transported_charge: float,
        p_h_1: Q_,
        p_h_2: Q_,
        ionic_strength_1: Q_,
        ionic_strength_2: Q_,
        delta_chi: Q_,
        temperature: Q_,
        p_mg_1: Q_ = default_pMg,
        p_mg_2: Q_ = default_pMg,
    ) -> ureg.Measurement:
        r"""Calculate the transformed energies of multi-compartmental reactions.

        Based on the equations from
        Harandsdottir et al. 2012 (https://doi.org/10.1016/j.bpj.2012.02.032)

        Parameters
        ----------
        reaction_1 : Reaction
            half-reaction associated to compartment 1
        reaction_2 : Reaction
            half-reaction associated to compartment 2
        transported_protons : float
            the total number of protons
            transported through the membrane
        transported_charge : float
            the total charge
            transported through the membrane
        p_h_1 : Quantity
            the pH in compartment 1
        p_h_2 : Quantity
            the pH in compartment 2
        ionic_strength_1 : Quantity
            the ionic strength in compartment 1
        ionic_strength_2 : Quantity
            the ionic strength in compartment 2
        delta_chi : Quantity
            electrostatic potential across the membrane
        temperature : Quantity
            temperature in Kalvin
        p_mg_1 : Quantity, optional
            the pMg in compartment 1 (Default value = 10)
        p_mg_2 : Quantity, optional
            the pMg in compartment 2 (Default value = 10)

        Returns
        -------
        standard_dg : Measurement
            the :math:`\Delta G'` in kJ/mol and standard error. to
            calculate the 95% confidence interval, multiply the error by 1.96

        """
        standard_dg, cov_dg = self.standard_dg_multi([reaction_1, reaction_2])

        # the contribution of the chemical reaction to the standard ΔG' is
        # the sum of the two half reactions and the variance is the sum
        # of all the values in cov_dg. For the uncertainty we take the square
        # root of the sum.

        uncertainty = np.sqrt(sum(cov_dg.flat))
        standard_dg = (standard_dg[0] + standard_dg[1]).plus_minus(uncertainty)

        # Here we calculate the transform (since the two reactions are in
        # different pH and I, we have to do it here):
        transform_dg = Q_(0.0, "kJ/mol")
        for compound, coeff in reaction_1.items(protons=False):
            transform_dg += coeff * compound.transform(
                p_h=p_h_1,
                ionic_strength=ionic_strength_1,
                temperature=temperature,
                p_mg=p_mg_1,
            )

        for compound, coeff in reaction_2.items(protons=False):
            transform_dg += coeff * compound.transform(
                p_h=p_h_2,
                ionic_strength=ionic_strength_2,
                temperature=temperature,
                p_mg=p_mg_2,
            )

        dg_protons = (
            transported_protons
            * R
            * temperature
            * np.log(10.0)
            * (p_h_2 - p_h_1)
        )

        dg_electrostatic = FARADAY * transported_charge * delta_chi

        standard_dg += transform_dg - dg_protons - dg_electrostatic
        return standard_dg.to("kJ/mol")

    def dg_analysis(self, reaction: Reaction) -> Iterable[Dict[str, object]]:
        r"""Analyse the weight of each observation to the :math:`\Delta G`.

        Parameters
        ----------
        reaction : Reaction
            the input Reaction object


        Returns
        -------
        list
            a list of reactions that contributed to the value of the
            :math:`\Delta G` estimation, with their weights and extra
            information

        """
        return self.preprocess.dg_analysis(reaction)

    def is_using_group_contribution(self, reaction: Reaction) -> bool:
        r"""Check if group contributions is needed for the :math:`\Delta G`.

        Parameters
        ----------
        reaction : Reaction
            the input Reaction object


        Returns
        -------
        bool
            True if the reaction require group contributions.

        """
        return self.preprocess.is_using_group_contribution(reaction)



"""module for component-contribution predictions."""
# The MIT License (MIT)
#
# Copyright (c) 2013 The Weizmann Institute of Science.
# Copyright (c) 2018 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
# Copyright (c) 2018 Institute for Molecular Systems Biology,
# ETH Zurich, Switzerland.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import logging
from collections import namedtuple
from io import BytesIO
from typing import BinaryIO, Union

import numpy as np
import pandas as pd

from . import ZENODO_DOI_PARAMETERS
from .diskcache import get_cached_file


logger = logging.getLogger(__name__)

PARAMETER_NAMES = (
    "train_b train_S train_w train_G "
    "dimensions "
    "dG0_rc dG0_gc dG0_cc "
    "V_rc V_gc V_inf MSE "
    "P_R_rc P_R_gc P_N_rc P_N_gc "
    "inv_S inv_GS inv_SWS inv_GSWGS "
)


class CCModelParameters(namedtuple("CCModelParameters", PARAMETER_NAMES)):
    """Container class for all Component Contribution parameters."""

    @staticmethod
    def from_zenodo(
        zenodo_doi: str = ZENODO_DOI_PARAMETERS,
    ) -> "CCModelParameters":
        """Get the CC parameters from Zenodo.

        Parameters
        ----------
        zenodo_doi : str
            The DOI for the Zenodo entry

        Returns
        -------
        CCModelParameters
            a collection of Component Contribution parameters.

        """
        params_file = get_cached_file(
            zenodo_doi=zenodo_doi, fname="cc_params.npz"
        )
        return CCModelParameters.from_npz(params_file)

    def to_npz(self, file: Union[str, BinaryIO]) -> None:
        """Save the parameters in NumPy uncompressed .npz format."""

        # convert the CCModelParameters object into a dictionary of NumPy
        # arrays. if one of the items is a pandas DataFrames, serialize it to 3
        # separate arrays (values, index, columns)
        param_dict = dict()
        for parameter_name in self._fields:
            parameter_value = self.__getattribute__(parameter_name)
            if type(parameter_value) == pd.DataFrame:
                param_dict[f"{parameter_name}_values"] = parameter_value.values
                param_dict[f"{parameter_name}_index"] = parameter_value.index
                param_dict[
                    f"{parameter_name}_columns"
                ] = parameter_value.columns
            else:
                param_dict[parameter_name] = parameter_value
        np.savez(file, **param_dict)

    @staticmethod
    def from_npz(file: Union[str, BytesIO]) -> "CCModelParameters":
        """Load the parameters from a NumPy uncompressed .npz file."""

        npzfile = np.load(file, allow_pickle=True)
        param_dict = dict(npzfile)

        # translate the serializes DataFrames back to the original form
        for df_name in ["dimensions", "MSE", "train_G", "train_S"]:
            param_dict[df_name] = pd.DataFrame(
                data=param_dict[f"{df_name}_values"],
                index=param_dict[f"{df_name}_index"],
                columns=param_dict[f"{df_name}_columns"],
            )
            del param_dict[f"{df_name}_index"]
            del param_dict[f"{df_name}_columns"]
            del param_dict[f"{df_name}_values"]
        return CCModelParameters(**param_dict)



"""A file with some Linear Algebra tools."""
# The MIT License (MIT)
#
# Copyright (c) 2018 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
from typing import Tuple

import numpy as np
from numpy.linalg import pinv
from scipy.linalg import qr, svd


class LINALG(object):
    """Some general linear algebra that's missing from scipy."""

    @staticmethod
    def invert_project(
        A: np.ndarray, tol: float = 1e-10
    ) -> Tuple[np.array, float, np.array, np.array]:
        """Use SVD to invert a matrix, while computing the subspace projections.

        Parameters
        ----------
        A : np.ndarray
            A 2D NumPy array
        tol : float, optional
            A threshold on the eigenvalues (below which are considered 0)
            (Default value = 1e-10)

        Returns
        -------
        tuple
            a 4-tuple of (inv(A), rank, P_R, P_N) where the P_R is the
            projection onto the range of A, and P_N is the projection onto the
            nullspace of A.T

        """
        M, N = A.shape

        U, s, Vh = svd(
            A,
            full_matrices=True,
            compute_uv=True,
            check_finite=True,
            lapack_driver="gesvd",
        )

        assert U.shape == (M, M), "SVD decomposition dimensions are wrong"
        assert Vh.shape == (N, N), "SVD decomposition dimensions are wrong"

        S = np.zeros((M, N))
        np.fill_diagonal(S, s)
        inv_A = Vh.T @ pinv(S) @ U.T

        rank = (S > tol).sum()
        P_R = U[:, :rank] @ U[:, :rank].T
        P_N = U[:, rank:] @ U[:, rank:].T

        assert inv_A.shape == (N, M), "Pseudoinverse dimensions are wrong"
        assert P_R.shape == (M, M), "Orthogonal projection dimensions are wrong"
        assert P_N.shape == (M, M), "Orthogonal projection dimensions are wrong"

        return inv_A, rank, P_R, P_N

    @staticmethod
    def row_uniq(A: np.ndarray) -> Tuple[np.array, np.array]:
        """Matrix Row Unique.

        A procedure usually performed before linear regression (i.e. solving
        Ax = y). If the matrix A contains repeating rows, it is advisable to
        combine all of them to one row, and the observed value corresponding
        to that row will be the average of the original observations.

        Parameters
        ----------
        A : np.ndarray
            A 2D NumPy array

        Returns
        -------
        tuple
            a 2-tuple (A_unique, P_row) where A_unique has the same
            number of columns as A, but with unique rows. P_row is a matrix that
            can be used to map the original rows to the ones in A_unique (all
            values in P_row are 0 or 1).

        """
        # convert the rows of A into tuples so we can compare them
        A_tuples = [tuple(A[i, :].flat) for i in range(A.shape[0])]
        A_unique = list(sorted(set(A_tuples), reverse=True))

        # create the projection matrix that maps the rows in A to rows in
        # A_unique
        P_col = np.zeros((len(A_unique), len(A_tuples)), dtype=float)

        for j, tup in enumerate(A_tuples):
            # find the indices of the unique row in A_unique which correspond
            # to this original row in A (represented as 'tup')
            i = A_unique.index(tup)
            P_col[i, j] = 1.0

        return np.array(A_unique, ndmin=2), P_col

    @staticmethod
    def col_uniq(A: np.array) -> Tuple[np.array, np.array]:
        """Matrix Column Unique.

        A procedure usually performed before linear regression (i.e. solving
        Ax = y). If the matrix A contains repeating columns, it is advisable to
        combine all of them to one column, and the observed value corresponding
        to that column will be the average of the original observations.

        Parameters
        ----------
        A : np.ndarray
            A 2D NumPy array

        Returns
        -------
        tuple
            a 2-tuple (A_unique, P_col) where A_unique has the same
            number of rows as A, but with unique columns. P_col is a matrix that
            can be used to map the original columns to the ones in A_unique (all
            values in P_col are 0 or 1).

        """
        A_unique, P_col = LINALG.row_uniq(A.T)
        return A_unique.T, P_col.T

    @staticmethod
    def qr_rank_deficient(A: np.array, tol: float = 1e-10) -> np.array:
        """QR recomposition for rank deficient matrices.

        Parameters
        ----------
        A : np.array
            A n-by-m matrix with rank r < min(n, m)
        tol : float, optional
            The tolerance below which an eigenvalue is considered 0.
            The default is 1e-10.

        Returns
        -------
        R : np.array
            A matrix of size r-by-m, that satisfies Q @ [R; 0] = A,
            where Q is unitary.

        """
        _, R, perm = qr(A, mode="economic", pivoting=True)
        r = sum(np.abs(np.diag(R)) > tol)

        # invert the permutation in order to apply it on M and return
        # the columns to the original order as in A
        perm = perm.tolist()
        invperm = [perm.index(i) for i in range(len(perm))]

        # keep only the non-zero rows, and reorder the columns according to
        # the original ordering in A
        return R[:r, invperm]



"""Handles downloading an caching of files from Zenodo."""
# The MIT License (MIT)
#
# Copyright (c) 2013 The Weizmann Institute of Science.
# Copyright (c) 2018 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
# Copyright (c) 2018 Institute for Molecular Systems Biology,
# ETH Zurich, Switzerland.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import hashlib
import pathlib
from io import BytesIO
from typing import Dict, Optional

import appdirs
import asyncio
import diskcache
import httpx
import pyzenodo3
from packaging import version
from tenacity import retry, stop_after_attempt
from tqdm.asyncio import tqdm


zen = pyzenodo3.Zenodo()


async def download_from_url(url: str) -> BytesIO:
    """Download a file from a given URL using httpx.

    Parameters
    ----------
    url : str
        The URL address of the file.
    md5 : str, optional
        The MD5 checksum of the file, if given and the checksum doesn't match
        the downaloded file, an IOError is raised. The default is None.

    Returns
    -------
    BytesIO
        Containing the downloaded file.

    """
    data = BytesIO()
    client = httpx.AsyncClient()
    async with client.stream("GET", url) as response:
        total = int(response.headers["Content-Length"])
        md5 = response.headers["content-md5"]

        num_bytes = 0
        with tqdm(
            total=total, unit_scale=True, unit_divisor=1024, unit="B"
        ) as progress:
            async for chunk in response.aiter_bytes():
                data.write(chunk)
                progress.update(len(chunk))
                num_bytes += len(chunk)
        await client.aclose()

    if num_bytes < total:
        raise IOError(f"Failed to download file from {url}")

    data.seek(0)
    if hashlib.md5(data.read()).hexdigest() != md5:
        raise IOError(f"MD5 mismatch while trying to download file from {url}")

    data.seek(0)
    return data


async def _get_zenodo_files(zenodo_doi: str) -> Dict[str, BytesIO]:
    """Run the get_zenodo_files coroutine asynchronously."""
    rec = zen.find_record_by_doi(zenodo_doi)
    fnames = [d["key"] for d in rec.data["files"]]
    urls = [d["links"]["self"] for d in rec.data["files"]]
    tasks = [download_from_url(url) for url in urls]
    data_streams = await asyncio.gather(*tasks)
    return dict(zip(fnames, data_streams))


@retry(stop=stop_after_attempt(3))
def get_zenodo_files(zenodo_doi: str) -> Dict[str, BytesIO]:
    """Download all the files stored in Zenodo (under a specific DOI).

    Parameters
    ----------
    zenodo_doi : str
        the DOI of the Zenodo entry.

    Returns
    -------
    Dict
        the dictionary with file names as keys, and the file contents as
        values.

    """
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    return loop.run_until_complete(_get_zenodo_files(zenodo_doi))


@retry(stop=stop_after_attempt(3))
def get_zenodo_version(zenodo_doi: str) -> Optional[str]:
    """Download all the files stored in Zenodo (under a specific DOI).

    Parameters
    ----------
    zenodo_doi : str
        the DOI of the Zenodo entry.

    Returns
    -------
    str
        latest version of the Zenodo entry.

    """
    try:
        rec = zen.find_record_by_doi(zenodo_doi)
    except pyzenodo3.base.requests.exceptions.ConnectionError:
        return None

    return rec.data["metadata"]["version"]


def get_cached_file(zenodo_doi: str, fname: str) -> BytesIO:
    """Get data from a file stored in Zenodo (or from cache, if available).

    Parameters
    ----------
    zenodo_doi : str
        the DOI of the Zenodo entry.
    fname : str
        the filename to fetch.

    Returns
    -------
    DataFrame
        the data contained in the file.

    """
    cache_directory = pathlib.Path(
        appdirs.user_cache_dir(appname="equilibrator")
    )
    online_version = get_zenodo_version(zenodo_doi)

    with diskcache.Cache(cache_directory) as cache:
        cached_version = cache.get(zenodo_doi + "/VERSION", None)

        if online_version is None:
            if cached_version is None:
                raise KeyError(
                    "Zenodo is unreachable and there is no local cached "
                    f"version for {zenodo_doi}."
                )
            # We have no connection to Zeonodo, so we use whatever version
            # that is cached.
            return cache[zenodo_doi][fname]

        if cached_version is not None:
            if version.parse(online_version) == version.parse(cached_version):
                return cache[zenodo_doi][fname]
            if version.parse(online_version) < version.parse(cached_version):
                raise KeyError(
                    f"The cached version of {zenodo_doi} ({cached_version}) "
                    f"is newer than the one online ({online_version})"
                )

        # If we are here, it means that there is no cached version or that
        # it is outdated. So, we download the files and update the version.
        dataframe_dict = get_zenodo_files(zenodo_doi)
        cache[zenodo_doi] = dataframe_dict
        cache[zenodo_doi + "/VERSION"] = online_version
        return cache[zenodo_doi][fname]



"""calculations needed for component-contribution predictions."""
# The MIT License (MIT)
#
# Copyright (c) 2013 The Weizmann Institute of Science.
# Copyright (c) 2018 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
# Copyright (c) 2018 Institute for Molecular Systems Biology,
# ETH Zurich, Switzerland.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import logging
from typing import Dict, Iterable, List, Optional, Tuple, Union

import numpy as np
from equilibrator_cache import Compound, Reaction

from . import CCModelParameters
from .linalg import LINALG


logger = logging.getLogger(__name__)


class Preprocessor(object):
    """A Component Contribution preprocessing class."""

    DEFAULT_RMSE_INF = 1e5

    def __init__(
        self, parameters: CCModelParameters, rmse_inf: float = DEFAULT_RMSE_INF
    ) -> None:
        """Create a GibbsEnergyPredictor object.

        Parameters
        ----------
        parameters : CCModelParameters
            all the parameters needed for running Component Contribution
            predictions.
        rmse_inf : float
            The MSE of the subspace that's not covered by Component Contribution
            which is set as an arbitrary high value. By default we set it to
            10^10 (kJ^2 / mol^2)

        """
        self._compound_ids = parameters.train_G.index.tolist()
        self.Nc = parameters.dimensions.at["Nc", "number"]

        # store the number of "real" groups, i.e. not including the "fake"
        # ones that are placeholders for non-decomposable compounds
        self.Ng = parameters.dimensions.at["Ng", "number"]

        self.S = parameters.train_S.values
        G = parameters.train_G.values

        self.RMSE_rc = np.sqrt(parameters.MSE.at["rc", "MSE"])
        self.RMSE_gc = np.sqrt(parameters.MSE.at["gc", "MSE"])
        self.RMSE_inf = rmse_inf

        self.mu = np.hstack([parameters.dG0_cc, parameters.dG0_gc[: self.Ng]])

        # pre-processing matrices
        self.L_rc = np.hstack(
            [
                parameters.inv_S @ parameters.P_R_rc,
                np.zeros((self.S.shape[1], self.Ng)),
            ]
        )
        self.L_gc = np.hstack(
            [
                parameters.inv_GS @ G.T @ parameters.P_N_rc,
                parameters.inv_GS[:, : self.Ng],
            ]
        )
        self.L_inf = np.hstack(
            [parameters.P_N_gc @ G.T, parameters.P_N_gc[:, : self.Ng]]
        )

        self.L_c = LINALG.qr_rank_deficient(
            np.vstack(
                [
                    self.RMSE_rc * self.L_rc,
                    self.RMSE_gc * self.L_gc,
                    self.RMSE_inf * self.L_inf,
                ]
            )
        )

    def get_compound_vector(self, compound: Compound) -> Union[np.array, None]:
        """Get the index of a compound in the original training data.

        Parameters
        ----------
        compound : Compound
            a Compound object


        Returns
        -------
        x
            the index of that compound, or -1 if it was not in the
            training list

        """

        try:
            # This compound is in the training set so we can use reactant
            # contributions for it
            i = self._compound_ids.index(compound.id)
            x = np.zeros(self.Nc + self.Ng, dtype=float)
            x[i] = 1
            return x
        except ValueError:
            if compound.group_vector:
                # This compound is not in the training set so must use group
                # contributions
                return np.hstack(
                    [np.zeros(self.Nc, dtype=float), compound.group_vector]
                )
            else:
                # This compound cannot be decomposed and therefore we cannot
                # have any estimate for it
                return None

    def get_compound_prediction(
        self, compound: Compound
    ) -> Tuple[Optional[float], Optional[np.ndarray]]:
        """Get the (mu, sigma) predictions of a compound's formation energy."""
        x = self.get_compound_vector(compound)
        if x is None:
            return None, None
        else:
            return self.mu @ x, self.L_c @ x

    def get_reaction_prediction(
        self, reaction: Reaction
    ) -> Tuple[float, np.ndarray, Dict[Compound, float]]:
        """Get the (mu, sigma) predictions of a reaction's energy.

        Parameters
        ----------
        reaction : Reaction
            the input Reaction object


        Returns
        -------
        mu : float
            the mean of the standard Gibbs energy estimate
        sigma : array
            a vector representing the square root of the covariance matrix
            (uncertainty)
        residual : dict
            the residual reaction in sparse notation (unknown and
            undecomposable reactants)

        """
        tot_mu = 0.0  # mean of the delta G estimate
        tot_sigma = np.zeros(self.L_c.shape[0])  # sqrt uncertainty vector
        residual = dict()

        for compound, coefficient in reaction.items(protons=False):
            mu, sigma = self.get_compound_prediction(compound)
            if mu is not None:
                tot_mu += coefficient * mu
                tot_sigma += coefficient * sigma
            else:
                residual[compound] = coefficient

        return tot_mu, tot_sigma, residual

    @staticmethod
    def _residuals_to_matrix(
        residuals: List[Dict[Compound, float]]
    ) -> np.ndarray:
        """Construct the residual stoichiometric matrix U.

        Parameters
        ----------
        residuals : List[dict]
            a list of dictionaries with each residual reaction in sparse
            notation (unknown and undecomposable reactants)

        Returns
        -------
        U : np.array
            a matrix with the square root of the covariance corresponding
            to the 'unknown' compounds (i.e. where the uncertainty is infinite)

        """

        # First, make an ordered list of the unknown-undecomposable compounds
        residual_compounds = set()
        for sparse in residuals:
            residual_compounds.update(sparse.keys())
        residual_compounds = sorted(residual_compounds)

        U = np.zeros((len(residual_compounds), len(residuals)))
        for i, sparse in enumerate(residuals):
            for cpd, coeff in sparse.items():
                j = residual_compounds.index(cpd)
                U[j, i] = coeff

        return U

    def get_reaction_prediction_multi(
        self, reactions: List[Reaction]
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Get the (mu, sigma) predictions of a reaction's energy.

        Parameters
        ----------
        reactions: List[Reaction] :
            a list of Reaction objects


        Returns
        -------
        mus : np.array
            the mean of the standard Gibbs energy estimates
        sigmas : np.array
            a matrix whose columns are the square root vectors (one for each
            reaction)
        U : np.array
            another matrix with the square root of the covariance corresponding
            to the 'unknown' compounds (i.e. where the uncertainty is infinite)

        """
        mus, sigmas, residuals = zip(
            *map(self.get_reaction_prediction, reactions)
        )
        mus = np.array(list(mus))
        sigmas = np.vstack(list(sigmas))
        U = self._residuals_to_matrix(list(residuals))

        return mus, sigmas, self.RMSE_inf * U

    def decompose_reaction(
        self, reaction: Reaction
    ) -> Tuple[np.ndarray, Dict[Compound, float]]:
        """Decompose a reaction.

        Parameters
        ----------
        reaction : Reaction
            the input Reaction object


        Returns
        -------
        x : ndarray
            the reaction vector (describing both RC and GC changes)
        residual : dict
            the residual reaction in sparse notation

        """
        x = np.zeros(self.Nc + self.Ng, dtype=float)
        residual = dict()
        for compound, coefficient in reaction.items(protons=False):
            _x = self.get_compound_vector(compound)
            if _x is None:
                residual[compound] = coefficient
            else:
                x += coefficient * _x

        return x, residual

    def dg_analysis(self, reaction: Reaction) -> Iterable[Dict[str, object]]:
        r"""Analyse the weight of each observation to the :math:`\Delta G`.

        Parameters
        ----------
        reaction : Reaction
            the input Reaction object


        Returns
        -------
        list
            a list of reactions that contributed to the value of the
            :math:`\Delta G` estimation, with their weights and extra
            information

        """
        x, residual = self.decompose_reaction(reaction)
        if residual:
            return []

        # dG0_cc = (x*G1 + x*G2 + g*G3)*b
        weights_rc = (self.L_rc @ x).round(5)
        weights_gc = (self.L_gc @ x).round(5)
        weights = abs(weights_rc) + abs(weights_gc)

        for j in reversed(np.argsort(weights)):
            if abs(weights[j]) < 1e-5:
                continue
            r = {
                i: self.S[i, j]
                for i in range(self.S.shape[0])
                if self.S[i, j] != 0
            }
            yield {
                "index": j,
                "w_rc": weights_rc[j],
                "w_gc": weights_gc[j],
                "reaction": r,
            }

    def is_using_group_contribution(self, reaction: Reaction) -> bool:
        r"""Check if group contributions is needed for the :math:`\Delta G`.

        Parameters
        ----------
        reaction : Reaction
            the input Reaction object


        Returns
        -------
        bool
            True if the reaction require group contributions.

        """
        x, residual = self.decompose_reaction(reaction)
        if residual:
            return False

        sum_w_gc = np.linalg.norm(self.L_gc @ x, 1)
        logging.info("sum(w_gc) = %.2g" % sum_w_gc)
        return sum_w_gc > 1e-5

    def get_reaction_prediction_orthogonal_dof(
        self, reactions: List[Reaction]
    ) -> np.ndarray:
        """Get the sigma_inf of a list reactions.

        Parameters
        ----------
        reactions: List[Reaction] :
            a list of Reaction objects


        Returns
        -------
        sigmas : np.array
            matrix whose columns (representing reactions) contain the degrees
            of freedom in the CC orthogonal space.

        """
        decompositions, residuals = zip(
            *map(self.decompose_reaction, reactions)
        )
        X = np.vstack(list(decompositions)).T
        U = self._residuals_to_matrix(list(residuals))
        return np.vstack([self.L_inf @ X, U])



"""Component Contribution Training script."""
# The MIT License (MIT)
#
# Copyright (c) 2013 The Weizmann Institute of Science.
# Copyright (c) 2018 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
# Copyright (c) 2018 Institute for Molecular Systems Biology,
# ETH Zurich, Switzerland.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import logging

import numpy as np
import pandas as pd

from . import CCModelParameters
from .linalg import LINALG
from .training_data import TrainingData


logger = logging.getLogger(__name__)


class Trainer(object):
    """A class for training the component contribution model."""

    @staticmethod
    def train(training_data: TrainingData) -> CCModelParameters:
        """Create a GibbsEnergyPredictor object.

        Parameters
        ----------
        training_data : TrainingData
            a TrainingData object containing all training data

        Returns
        -------
        CCModelParameters
            The parameters of the Component Contribution model.

        """
        G = Trainer.group_incidence_matrix(training_data)
        S = training_data.stoichiometric_matrix
        b = training_data.standard_dg
        w = training_data.weight

        # We need to convert the Compound objects in the index of S and G to
        # serializable primitives, in order to later store them in Quilt.
        # Therefore, we replace the Compound object with the Compound.id

        S.index = S.index.map(lambda c: c.id)
        G.index = G.index.map(lambda c: c.id)

        # The number of "real" groups (i.e. excluding the non-decomposable
        # compounds). This is needed to know the length of group vectors.
        number_of_groups = training_data.group_summary.shape[0]

        params = Trainer.train_from_matrices(S, G, b, w, number_of_groups)
        return params

    @staticmethod
    def group_incidence_matrix(training_data: TrainingData) -> pd.DataFrame:
        """Initialize G matrix.

        Decompose each of the compounds that has an InChI into groups,
        and save the decomposition as a row in the G matrix.
        Compounds that cannot be decomposed are added as new "groups",
        i.e. as extra columns to the matrix.

        training_data : TrainingData
            a TrainingData object containing all training data

        Returns
        -------
        pd.DataFrame
            The group incidence matrix as a pd.DataFrame.

        """
        non_decomposable_compounds = (
            training_data.non_decomposable_compounds.copy()
        )

        compound_to_gv = {}
        # decompose the compounds in the training_data and add to G
        for compound in training_data.decomposable_compounds:
            if compound.group_vector:
                compound_to_gv[compound] = compound.group_vector
            else:
                non_decomposable_compounds.add(compound)
                if compound.group_vector is None:
                    logger.error(
                        "The compound (%r) is not decomposed yet, but is "
                        "in the CC training dataset. Please run "
                        "'populate_cache 'decompose' and try again.",
                        compound,
                    )

        _group_df = training_data.group_summary

        group_data = [
            compound_to_gv.get(cpd, [0] * _group_df.shape[0])
            for cpd in training_data.compounds
        ]

        G = pd.DataFrame(
            index=training_data.compounds,
            columns=_group_df.full_name.tolist(),
            data=group_data,
            dtype=float,
        )

        for compound in non_decomposable_compounds:
            # add a new column corresponding to this non-decomposable group
            G[compound] = 0.0

            # place a single '1' for this compound group decomposition
            G.at[compound, compound] = 1.0

        return G

    @staticmethod
    def train_from_matrices(
        train_S: pd.DataFrame,
        train_G: pd.DataFrame,
        train_b: pd.Series,
        train_w: pd.Series,
        number_of_groups: int,
    ) -> CCModelParameters:
        """Estimate standard Gibbs energies of formation.

        Parameters
        ----------
        train_S : pd.DataFrame
            training data stoichiometric matrix
        train_G : pd.DataFrame
            group incidence matrix
        train_b : pd.Series
            training data reverse transformed dG'0 values
        train_w : pd.Series
            weights
        number_of_groups : int
            the number of "real" groups

        Returns
        -------
        CCModelParameters
            Collection of Component Contribution parameters trained using
            the input data.

        """
        assert (train_G.index == train_S.index).all()
        assert (train_b.index == train_S.columns).all()
        assert (train_w.index == train_S.columns).all()

        S = train_S.values
        G = train_G.values
        b = np.array([dg.m_as("kJ/mol") for dg in train_b.tolist()], ndmin=1)
        w = np.array(train_w.tolist(), ndmin=1)

        # Apply weighing
        W = np.diag(w.ravel())
        GS = G.T @ S

        # Linear regression for the reactant layer (aka RC)
        inv_S, r_rc, P_R_rc, P_N_rc = LINALG.invert_project(S @ W)

        # Linear regression for the group layer (aka GC)
        inv_GS, r_gc, P_R_gc, P_N_gc = LINALG.invert_project(GS @ W)

        # calculate the group contributions
        dG0_gc = inv_GS.T @ W @ b

        # Calculate the contributions in the stoichiometric space
        dG0_rc = inv_S.T @ W @ b
        dG0_cc = P_R_rc @ dG0_rc + P_N_rc @ G @ dG0_gc

        # Calculate the residual error (unweighted squared error divided
        # by N - rank)
        e_rc = S.T @ dG0_rc - b
        MSE_rc = (e_rc.T @ W @ e_rc) / (S.shape[1] - r_rc)

        e_gc = GS.T @ dG0_gc - b
        MSE_gc = (e_gc.T @ W @ e_gc) / (S.shape[1] - r_gc)

        # Calculate the MSE of GC residuals for all reactions in ker(G).
        # This will help later to give an estimate of the uncertainty for such
        # reactions, which otherwise would have a 0 uncertainty in the GC
        # method.
        kerG_inds = list(np.where(np.all(GS == 0, 0))[0].flat)

        e_kerG = e_gc[kerG_inds]
        MSE_kerG = (e_kerG.T @ e_kerG) / len(kerG_inds)

        # Calculate the uncertainty covariance matrices
        inv_SWS, _, _, _ = LINALG.invert_project(S @ W @ S.T)
        inv_GSWGS, _, _, _ = LINALG.invert_project(GS @ W @ GS.T)

        V_rc = P_R_rc @ inv_SWS @ P_R_rc
        V_gc = P_N_rc @ G @ inv_GSWGS @ G.T @ P_N_rc
        V_inf = P_N_rc @ G @ P_N_gc @ G.T @ P_N_rc

        MSE = pd.DataFrame(
            index=["rc", "gc", "kerG"],
            columns=["MSE"],
            data=[MSE_rc, MSE_gc, MSE_kerG],
        )

        dimensions = pd.DataFrame(
            index=["Nc", "Ng", "Ng_full"],
            columns=["number"],
            data=[train_G.shape[0], number_of_groups, train_G.shape[1]],
        )

        # Put all the calculated data in 'params' for the sake of debugging
        return CCModelParameters(
            train_b=b,
            train_S=train_S,
            train_w=w,
            train_G=train_G,
            dimensions=dimensions,
            dG0_rc=dG0_rc,
            dG0_gc=dG0_gc,
            dG0_cc=dG0_cc,
            V_rc=V_rc,
            V_gc=V_gc,
            V_inf=V_inf,
            MSE=MSE,
            P_R_rc=P_R_rc,
            P_R_gc=P_R_gc,
            P_N_rc=P_N_rc,
            P_N_gc=P_N_gc,
            inv_S=inv_S,
            inv_GS=inv_GS,
            inv_SWS=inv_SWS,
            inv_GSWGS=inv_GSWGS,
        )
